'''
Created on Dec 11, 2019

@author: radek
'''

Set_C = set

class FSM_C(object):
  #-----------------
  """A simple model of Moore FSM dedicated to active learning."""
  
  def __init__(self, startState, transitions, output):
    #=================================================
    """Constructor."""
    self.startState  = startState
    self.transitions = transitions
    self.output      = output

  def InputAlphabet(self):
    #=====================
    """Returns all possible input symbols."""
    iChars = Set_C()
    for events in self.transitions.values():
      iChars |= Set_C(events)
    return sorted(list(iChars))
  
  def ApplySequence(self, inputSequence):
    #====================================
    """Returns the last state and the associated output symbol of a state sequence 
    if the input 'inputSequence' is applied to the fsm starting from the starting state."""
    cState = self.startState
    for iChar in inputSequence:
      cState = self.transitions[cState][iChar]
    return cState, self.output[cState]
  
  def Print(self):
    #=============
    """Prints the specification of the fsm (the starting state, the transition table, and
    the output table."""
    print('FSM', self.startState)
    for s, row in self.transitions.items():
      print(' t ', s, row)
    for s, outChar in self.output.items():
      print(' o ', s, outChar)

  def Tests(self, k):
    #================
    """Creates a test set composed from the state cover, the transition cover, 
    and the transition cover prolongated with one step further."""
    cover = {(): self.output[self.startState]}
    inChars = self.InputAlphabet()
    # State cover
    stateCover = {self.startState: ()}
    queue = [self.startState]
    while queue:
      s = queue.pop(0)
      for e in inChars:
        sNext = self.transitions[s][e]
        if sNext not in stateCover:
          test = stateCover[s] + (e,)
          cover[test] = self.output[sNext]
          stateCover[sNext] = test
          queue.append(sNext)
    # Transition cover
    for srcS in self.transitions:
      for e in inChars:
        dstS = self.transitions[srcS][e]
        test = stateCover[srcS] + (e,)
        cover[test] = self.output[dstS]
    # Missing verification of states!!!
    tCover = list(cover.keys())
    for t in tCover:
      for e in inChars:
        tExt = t+ (e,)
        dstS, dstO = self.ApplySequence(tExt)
        cover[tExt] = dstO
    return cover
    
class Teacher_C(object):
  #---------------------
  """An interface for FSM active learning covering communication with an unknown system."""
  
  def __init__(self, fsm):
    #=====================
    """Constructor. The unknown system is assumed to behave as a Moore machine."""
    self.fsm = fsm

  def InputAlphabet(self):
    #=====================
    """Returns the input alphabet of the attached unknown system."""
    return self.fsm.InputAlphabet()
  
  def Output(self, inputSequence):
    #=============================
    """Return the last output symbol if the input sequence 'inputSequence' is applied to
    the attached system from its starting state."""
    fState, fOut = self.fsm.ApplySequence(inputSequence)
    return fOut

  def GetCounterExample(self, dfa, k):
    #=================================
    """Returns a shortest sequence for which the attached unknown system and the verified
    FSM model 'dfa'."""
    for t, outChar in sorted(dfa.Tests(k).items(), key = lambda x: len(x[0])):
      s1, o1 = dfa.ApplySequence(t)
      s2, o2 = self.fsm.ApplySequence(t)
      if 1: print('Test', t, ':', s1, o1, '?', s2, o2)
      if o1 != o2:
        return (t,)
    return None

def GetATeacher():
  #===============
  """Returns an exemplar teacher attached to an unknown system."""
  fsmT = {
    'S': {'a': 'X', 'b': 'Y'},
    'X': {'a': 'Y', 'b': 'S'},
    'Y': {'a': 'S', 'b': 'X'},
  }
  fsmOut = {'S': 0, 'X': 1, 'Y': 0}  
  
  return Teacher_C(FSM_C('S', fsmT, fsmOut))
  
def TestFSM():
  #===========
  teacher = GetATeacher()
    
if __name__ == '__main__':
  if 1: TestFSM()