'''
Created on Dec 11, 2019

@author: radek
'''

from FSM import FSM_C, GetATeacher

class LStarModel_C(object):
  #------------------------

  eLambda = () # Empty sequence
  iterationLimit = 100
  
  def __init__(self, teacher):
    #=========================
    """Constructor."""
    self.teacher  = teacher                       # A teacher that answers queries and provides counterexamples
    self.alphabet = [(a,) for a in self.teacher.InputAlphabet()]  # Input alphabet
    self.sEntries = [self.eLambda]                # the top part of the table --> states, i.e. S part
    self.eEntries = [self.eLambda]                # table columns --> distinguishing set
    self.table = {                                # observation table: access key -> a row of output for all eEntries
      self.eLambda: {
        self.eLambda: self.teacher.Output(self.eLambda), # Init S part
        }}
    self.ExtendTable() # Initialize S.A part of the table

    if 1:
      print('input alphabet', self.alphabet)

  def PrintTable(self):
    #==================
    """Prints the current state of the observation table."""
    print('Table:  ')
    print('  S =', self.sEntries)
    print('  E =', self.eEntries)
    sKeys = sorted(list(self.table.keys()))
    eKeys = sorted(self.eEntries[:])
    for s in sKeys:
      print('  ', '*' if s in self.sEntries else ' ', '%-15s:' % str(s), end = ' ')
      for e in eKeys:
        if e not in self.table[s]:
          print('FAILURE: table', repr(s), repr(e))
          print('   ', self.table[s].keys())
        print(e, '=', self.table[s][e], end = ' ')
      print()

  def AddRow(self, ns):
    #==================
    """Ensures that the row with 'ns' access key has all outputs for all current eEntries fulfilled."""
    if ns not in self.table:
      self.table[ns] = {}
    for e in self.eEntries:
      if e not in self.table[ns]:
        if 1: print('Adding', ns, e, ns+e)
        self.table[ns][e] = self.teacher.Output(ns+e)
    
  def ExtendTable(self):
    #===================
    """Ensures that the observation table has both S and S.A part fully fulfiled."""
    for s in self.sEntries:
      self.AddRow(s)
      for a in self.alphabet:
        self.AddRow(s+a)
          
  def StateCode(self, sEntry):
    #=========================
    """Returns the state code Id used in DFA model. The state code is a tuple of outputs
    for the current order set of eEntries."""
    # L* state id -> FSM state id
    row = self.table[sEntry]
    return tuple([row[e] for e in self.eEntries])

  def AddCounterExample(self, ce):
    #=============================
    """Adds all prefixes of the counter example input sequence to the observation table.
    It also ensures that the observation table is fully fulfiled."""
    if 1: print('counterexample', ce)
    for ceLength in range(len(ce)):
      sPrefix = ce[:ceLength]
      if 0: print('sPrefix', sPrefix)
      if sPrefix not in self.sEntries:
        self.sEntries.append(sPrefix)
      for a in [self.eLambda]+self.alphabet:
        self.AddRow(sPrefix + a)
    
  def RowDifference(self, row1, row2):
    #=================================
    """Returns an symbol of eEntries which distinguishes
    two given rows 'row1' and 'row2' of the observation table."""
    for eSymbol in self.eEntries:
      if row1[eSymbol] != row2[eSymbol]:
        return eSymbol
    return None
  
  def InconsistentTail(self):
    #========================
    """If the observation table is not consistent then it returns the event and
    a distinguishing sequence (a new eTable entry) otherwise None."""
    if 1: print('InconsistentTail', self.sEntries)
    #TODO !!!
    if 1:
      a = self.alphabet[0]
      eDiff = self.eEntries[0]
      return a, eDiff
    return None
  
  def UnclosedTail(self):
    #====================
    """Returns None if the self.table is closed.
    The function returns s in self.sEntries and the input symbol 'a'
    if the table is unclosed. The access code 's+a' refers to a row of the S.A part
    of the observation table which output configuration is not registered in the S part."""
    if 1: print('UnclosedTail S=', self.sEntries)
    #TODO !!!
    if 1:
      s = self.sEntries[0]
      a = self.alphabet[0]
      return s, a
    return None


  def ConjectureDFA(self):
    #=====================
    """Returns a conjecture FSM derived from self.table.
    It is recommended to used self.StateCode(..) for state representation."""
    initState    = self.StateCode(self.eLambda)
    fTrans       = {}
    fOut         = {}
    #  TODO !!!
    return FSM_C(initState, fTrans, fOut)
    
  def LearnDFA(self):
    #================
    """Implements the active learning algorithm. The observation table is assumed
    to be initialized by the constructor. The fuction returns the learned instance
    of the FSM_C class representing a Moore finite state machine."""
    #TODO !!!
    dfa = None
    for i in range(self.iterationLimit):
      self.PrintTable()
      dfa = self.ConjectureDFA()
    return dfa

def TestLStarLearning():
  #=====================
  """A simple example demonstrating active learning of finite state machines using a teacher."""
  teacher = GetATeacher()
  lsm = LStarModel_C(teacher)
  dfa = lsm.LearnDFA()
  dfa.Print()
  
  
if __name__ == '__main__':
  if 1: TestLStarLearning()
  
